#partoffer.py
#Class encapsulating an Octopart PartOffer response, exposing the 
#price breaks 

from prices import PriceList
import utilities

class PartOffer:
	def __init__(self, response):
		utilities.assert_class(response, 'PartOffer')

		self.seller_name = response['seller']['name']
		self.seller_uid = response['seller']['uid']
		self.packaging = response['packaging']
		self.invalid = False

		try:
			self.breaks = PriceList(response['prices'])
		except ValueError, e:
			self.breaks = None
			self.invalid = True

		self.in_stock_quant = response['in_stock_quantity']


	def best_price(self, quantity):
		if self.invalid:
			return None

		return self.breaks.unit_price(quantity)


import proxy12
from pymomo.commander.exceptions import *
from pymomo.commander.types import *
from pymomo.commander.cmdstream import *
from pymomo.utilities.console import ProgressBar
import struct
from intelhex import IntelHex
from time import sleep

class MultiSensorModule (proxy12.MIB12ProxyObject):
	ranges = {'offset': set(range(0, 256)), 'gain1': set(range(0, 128)), 'gain2':set(range(0,8)), 'delay': set(range(1, 256)), 
				'invert': set(['yes', 'no']), 'select':set(['current', 'differential'])}
	mapper = {'yes': 1, 'no': 0, 'current': 1, 'differential': 0}
	ids = {'offset': 0, 'gain1': 1, 'gain2': 2, 'select': 3, 'invert': 4, 'delay': 5}
	integral = set(['offset', 'gain1', 'gain2', 'delay'])

	def __init__(self, stream, addr):
		super(MultiSensorModule, self).__init__(stream, addr)
		self.name = 'MultiSensor Module'

	def set_parameter(self, name, value):
		if name not in self.ranges:
			raise KeyError("Invalid parameter options are: %s" % str(self.ranges.keys()))

		if name in self.integral:
			value = int(value)

		valid = self.ranges[name]
		if value not in valid:
			raise ValueError("Invalid parameter value for key %s" % name)

		if not isinstance(value, int):
			if value not in self.mapper:
				raise ValueError("Invalid parameter value for key %s, cannot be mapped to a known integer" % name)

			value = self.mapper[value]

		self.rpc(20, 3, self.ids[name], value)

	def read_voltage(self):
		res = self.rpc(20, 5, result_type=(1,False))
		return res['ints'][0]

	def acquire_pulses(self):
		self.rpc(20, 6)

	def read_pulses(self):
		res = self.rpc(20, 7, result_type=(1, False))
		return res['ints'][0]


#pyeagle.py

#Routines that encapsulate the use of EAGLE to automatically generate CAM files
#and assembly drawings.

import shutil
import subprocess
import os.path
import os
from pybom.board import Board
import zipfile
import utilities.config

settings = utilities.config.ConfigFile('settings')

def execute_eagle(args):
	eagle = 'eagle'

	with open(os.devnull, 'wb') as DEVNULL:
		subprocess.check_call([eagle] + args, stdout=DEVNULL, stderr=DEVNULL)

def export_image(board, output, layers):
	process_section(board, output, 'PS', layers)

def process_section(board, output, type, layers):
	"""
	Call the EAGLE program to process a board file and produce a 
	CAM output with the given layers on it.

	@param type can be either "gerber" or "excellon" to generate
	either Gerber 274X files or excellon drill files
	@param output complete path to the output file location
	@param layers the EAGLE layers to include in this CAM file
	@param board the complete path to the board file
	"""

	#Options from EAGLE's built in gerber file for producing 2 layer boards
	#and argument names from EAGLE -? command
	args = ['-X', '-f+', '-c+', '-O+']

	if type == 'gerber':
		args.append('-dGERBER_RS274X')
		remext = '.gpi'
	elif type == 'excellon':
		args.append('-dEXCELLON')
		remext = '.dri'
	elif type == 'PS':
		args.append('-dPS')
		args.append('-h11')
		args.append('-w7.75')
		args.append('-s2.5')
	else:
		raise ValueError("Invalid type specified (must be gerber or excellon), was: %s" % type)

	args.append('-o%s' % output)

	args.append(board)

	for layer in layers:
		args.append(layer)

	execute_eagle(args)

	#if it's gerber, remove gpi file
	#if it's excellon, remove drd file
	if type is not 'PS':
		(base, ext) = os.path.splitext(output)
		remfile = base + remext

		os.remove(remfile)

def process_2layer(board, output_dir, basename, paste=False):
	"""
	Process the eagle file board specified to produce the correct gerber files for
	fabrication and assembly.  All output files will have the same basename with 
	different extensions
	"""

	top_silk = os.path.join(output_dir, basename + '.plc')
	top_copper = os.path.join(output_dir, basename + '.cmp')
	bot_copper = os.path.join(output_dir, basename + '.sol')
	bot_mask = os.path.join(output_dir, basename + '.sts')
	top_mask = os.path.join(output_dir, basename + '.stc')
	drill = os.path.join(output_dir, basename + '.drd')
	top_cream = os.path.join(output_dir, basename + '.crm')

	#Make sure the output dir exists
	ensure_dir_exists(output_dir)

	process_section(board, top_copper, 'gerber', ['Top', 'Pads', 'Vias'])
	process_section(board, bot_copper, 'gerber', ['Bottom', 'Pads', 'Vias'])
	process_section(board, top_silk, 'gerber', ['Dimension', 'tPlace', 'tNames'])
	process_section(board, top_mask, 'gerber', ['tStop'])
	process_section(board, bot_mask, 'gerber', ['bStop'])
	process_section(board, drill, 'excellon', ['Drills', 'Holes'])

	if paste:
		process_section(board, top_cream, 'gerber', ['tCream'])

def ensure_dir_exists(output_dir):
	if not os.path.isdir(output_dir):
		os.makedirs(output_dir)

def create_readme(output_dir, basename, brd_obj, paste=False):
	with open(os.path.join(output_dir, 'README.txt'), "w") as f:
		f.write('WellDone\n')
		f.write("PCB Fabrication Files\n")
		f.write("Name: %s\n" % brd_obj.partname)
		f.write("Revision: %s\n" % brd_obj.revision)
		f.write("Dimensions: %sx%s inches\n" % (brd_obj.width, brd_obj.height))
		f.write("Contact: Tim Burke <tim@welldone.org>\n\n")
		f.write("Folder Contents:\n")
		f.write('%s: Top Silkscreen\n' % (basename+'.plc'))
		f.write('%s: Top Copper\n' % (basename+'.cmp'))
		f.write('%s: Top Soldermask\n' % (basename+'.stc'))
		f.write('%s: Bottom Soldermask\n' % (basename+'.sts'))
		f.write('%s: Bottom Copper\n' % (basename+'.sol'))

		if paste:
			f.write('%s: Top Cream\n' % (basename+'.crm'))

		f.write('%s: Excellon Drill File\n' % (basename+'.drd'))

def build_assembly_drawing(board, output):
	"""
	Create an assembly drawing for this board.  File created will by a PS file
	"""

	export_image(board, output, ['tPlace', 'tNames', 'tDocu', 'Document', 'Reference','Dimension'])

def build_production(board, output_dir, paste=False):
	"""
	Build the set of production files associated with this EAGLE board file.
	Directory structure will be:
	output_dir
		- fabrication
			+ CAM files
			+ README
		- basename_fab.zip
		- assembly
			+ basename_bom.csv
			+ basename_drawing.ps
	"""

	board_obj = Board.FromEagle(board)
		
	basename = board_obj.partname

	fab_dir = os.path.join(output_dir, 'fabrication')
	ass_dir = os.path.join(output_dir, 'assembly')

	#Ensure old fabrication and assembly files are removed
	if os.path.isdir(fab_dir):
		shutil.rmtree(fab_dir)
	if os.path.isdir(ass_dir):
		shutil.rmtree(ass_dir)

	#Create fabrication files
	process_2layer(board, fab_dir, basename, paste=paste)
	create_readme(fab_dir, basename, board_obj, paste=paste)
	zipfab(fab_dir, os.path.join(output_dir, basename + '_fab'))

	#Create assembly files
	ensure_dir_exists(ass_dir)
	build_assembly_drawing(board, os.path.join(ass_dir, basename + '_drawing.ps'))

	#Build a BOM for each assembly variant
	for var in board_obj.variants.keys():
		board_obj.export_bom(var,os.path.join(ass_dir, basename + "_" + var + "_bom.csv"))

def zipfab(path, output):
	"""
	Create a zipfile of the direction path with the name output.zip that will expand into a directory 
	with the same name as output containing all of the files in path. 
	"""
	zip = zipfile.ZipFile(output+'.zip', 'w', zipfile.ZIP_DEFLATED)
	for root, dirs, files in os.walk(path):
		for file in files:
			zip.write(os.path.join(root, file), os.path.join(os.path.basename(output), file))

	zip.close()



class Package:
	"""
	A pcb component package containing the number of pads and pins in the package
	as well as its name.
	"""

	def __init__(self, element):
		if element.tag != 'package':
			raise TypeError('You must pass an ElementTree element to Package.__init__')

		self.name = element.get('name', default="Unknown Name")

		pads = element.findall('./smd')
		self.num_pads = len(pads)

		pins = element.findall('./pad')
		self.num_pins = len(pins)


#octopart.py
#Wrapper arround Octopart REST API

import os
import json
import urllib
import pprint
from physicalpart import PhysicalPart
import sys
from partcache import PartCache

from pymomo.utilities.paths import MomoPaths

class Octopart:
	URL = 'http://octopart.com/api/v3/'

	def __init__(self, key=None):
		if key is not None:
			self.api_key = key
		elif 'OCTOPART_KEY' in os.environ:
			self.api_key = os.environ['OCTOPART_KEY']
		else:
			raise ValueError('You need to either specify an Octopart API Key or have OCTOPART_KEY set in the shell')

		self.cache = PartCache()

	def _build_call(self, method, args):
		"""
		Build a URL call with arg=val pairs including the apikey to the url specifying the method
		given.
		"""

		url = Octopart.URL + method + '?'

		out_args = []
		out_args.extend(args)
		out_args.append(('apikey', self.api_key))
		out_strings = map(lambda x: urllib.quote(x[0])+'='+urllib.quote(x[1]), out_args)

		query_string = "&".join(out_strings)

		return url + query_string

	def _call(self, url):
		data = urllib.urlopen(url).read()
		response = json.loads(data)

		return response

	def _call_method(self, method, args):
		url = self._build_call(method, args)
		return self._call(url)

	def match_identifiers(self, skus):
		matched = {}
		unmatched = []
		uncached_skus = []

		#Check if we have these parts in our cache
		for sku in skus:
			id = sku.build_reference()
			part = self.cache.get(id)

			if part is None:
				uncached_skus.append(sku)
			else:
				matched[id] = part

		#Otherwise fetch from Octopart and cache the results
		for i in xrange(0, len(uncached_skus), 20):
			req = uncached_skus[i:i+20]
			queries = map(lambda x: x.build_query(), req)

			query = json.dumps(queries)

			resp = self._call_method('parts/match',[('queries',query)])

			for result in resp['results']:
				ref = result['reference']

				if len(result['items']) == 0:
					unmatched.append(ref)
					print "Did not match %s" % ref
					continue

				if len(result['items']) > 1:
					print "Found %d parts for %s" % (len(result['items']), ref)
				
				part = PhysicalPart(result['items'][0])
				self.cache.set(ref, part)

				matched[ref] = part 

		return matched


#External API functions from this package

from annotate import param, returns, context, finalizer, takes_cmdline, annotated
from typeinfo import is_known_type, is_known_format, convert_to_type, format_value, get_type_size, inject_type, load_external_types


#typeinfo.py
#Basic routines for converting information from string or other binary
#formats to python types and for displaying those types in supported 
#formats

from pymomo.exceptions import *
import annotate
import types
import os.path
import imp

def convert_to_type(value, type, **kwargs):
	"""
	Convert value to type 'type'

	If the conversion routine takes various kwargs to 
	modify the conversion process, **kwargs is passed
	through to the underlying conversion function
	"""

	if not is_known_type(type):
		raise ArgumentError("type is not known to type system", type=type)

	typeobj = getattr(types, type)

	conv = typeobj.convert(value, **kwargs)
	return conv

def get_type_size(type):
	"""
	Get the size of this type for converting a hex string to the
	type. Return 0 if the size is not known.
	"""

	if not is_known_type(type):
		raise ArgumentError("type is not known to type system", type=type)

	typeobj = getattr(types, type)

	if hasattr(typeobj, 'size'):
		return typeobj.size()

	return 0

def format_value(value, type, format=None, **kwargs):
	"""
	Convert value to type and format it as a string

	type must be a known type in the type system and format,
	if given, must specify a valid formatting option for the
	specified type.
	"""

	typed_val = convert_to_type(value, type, **kwargs)
	typeobj = getattr(types, type)

	#Allow types to specify default formatting functions as 'default_formatter'
	#otherwise if not format is specified, just convert the value to a string
	if format is None:
		if hasattr(typeobj, 'default_formatter'):
			format_func = getattr(typeobj, 'default_formatter')
			return format_func(typed_val, **kwargs)

		return str(typed_val)

	formatter = "format_%s" % str(format)
	if not hasattr(typeobj, formatter):
		raise ArgumentError("Unknown format for type", type=type, format=format, formatter_function=formatter)

	format_func = getattr(typeobj, formatter)
	return format_func(typed_val, **kwargs)

def is_known_type(type):
	"""
	Check if type is known to the type system.

	Returns boolean indicating if type is known.
	"""

	if not isinstance(type, basestring):
		raise ArgumentError("type must be a string naming a known type", type=type)

	if not hasattr(types, type):
		return False

	return True

def is_known_format(type, format):
	"""
	Check if format is known for given type.

	Returns boolean indicating if format is valid for the specified type.
	"""

	if not is_known_type(type):
		return False

	typeobj = getattr(types, type)
	formatter = "format_%s" % str(format)
	if not hasattr(typeobj, formatter):
		return False

	return True

def inject_type(name, typeobj):
	"""
	Given a module-like object that defines a type, add it to our type system so that
	it can be used with the momo tool and with other annotated API functions.
	"""

	#TODO add checking each type for the minimum required content like a default formatter,
	#a conversion function, etc.

	if is_known_type(name):
		raise ArgumentError("attempting to inject a type that is already defined", type=name)

	setattr(types, name, typeobj)

def load_external_types(path, verbose=False):
	"""
	Given a path to a python package or module, load that module, search for all defined variables
	inside of it that do not start with _ or __ and inject them into the type system.  If any of the
	types cannot be injected, silently ignore them unless verbose is True.  If path points to a module
	it should not contain the trailing .py since this is added automatically by the python import system
	"""

	d,p = os.path.split(path)

	try:
		fileobj,pathname,description = imp.find_module(p, [d])
		mod = imp.load_module(p, fileobj, pathname, description)
	except ImportError:
		raise ArgumentError("could not import module in order to load external types", module_path=path, parent_directory=p, module_name=p)

	for name in filter(lambda x: not x.startswith('_'), dir(mod)):
		typeobj = getattr(mod, name)
		inject_type(name, typeobj)

	#TODO add checking for types that could not be injected and report them



#identifier.py

class PartIdentifier:
	def __init__(self, part=None, digipn=None):

		if digipn is not None:
			self.sku = digipn
			self.mpn = None
			self.manu = None
		else:
			self.sku = part.digipn
			self.mpn = part.mpn
			self.manu = part.manu

	def build_reference(self):
		if self.sku:
			return "__SKU__" + self.sku

		else:
			return "__MPN__" + str(self.mpn)

	def build_query(self):
		if self.sku is not None:
			return {'sku': self.sku, 'reference': self.build_reference()}

		else:
			return {'mpn': self.mpn, 'reference': self.build_reference()}



#partcache.py
#A simple expiring cache built on top of zodb

import sys
import os.path

from pymomo.utilities.paths import MomoPaths
from ZODB.FileStorage import FileStorage
from ZODB.DB import DB
from time import time
import transaction

class PartCache(object):
	URL = 'http://octopart.com/api/v3/'
	CachePath = os.path.join(MomoPaths().config, 'part_lookup.cache')
	DefaultExpiry = 10*24*60*60		#Default expiration date is 10 days

	def __init__(self):
		storage = FileStorage(PartCache.CachePath)
		self.db = DB(storage)
		self.connection = self.db.open()
		self.root = self.connection.root()

	def _check_valid(self, obj):
		exp = obj['expiration']

		cur = time()

		if exp > cur:
			return True

		return False

	def get(self, id):
		if id not in self.root:
			return None

		obj = self.root[id]

		if self._check_valid(obj):
			obj['last_used'] = time()
			self.root[id] = obj
			transaction.commit()
			return obj['data']

		#If it's expired, delete it
		del self.root[id]
		return None

	def set(self, id, obj, expire=None):
		if expire is None:
			expire = PartCache.DefaultExpiry

		entry = {}
		entry['expiration'] = time() + expire
		entry['data'] = obj
		entry['last_used'] = time()

		self.root[id] = entry
		transaction.commit() 


#part.py

import re
import reference

known_types = {
	"C": "capacitor",
	"R": "resistor",
	"D": "diode",
	"JP": "connector",
	"U": "ic",
	"XTAL": "crystal",
	"L": "inductor",
	"LED": "led"
}

class_table = {
	""
}

class Part:
	"""
	An electronic component.
	"""

	ref = reference.PCBReferenceLibrary()

	@staticmethod
	def FromBoardElement(elem, variant, packages):
		"""
		Create a Part object from an element in an EAGLE board file. 
		pn_attrib specifies the attribute name that contains the correct
		manufacturer or distributor part number for this part.

		@TODO:
			- parse pn_attrib to determine whether its a distributor pn or a manu pn
			and set attributes accordingly
		"""

		pkg = elem.get('package', 'Unknown Package')
		pkg_info = None

		if pkg in packages:
			pkg_info = packages[pkg]

		name = elem.get('name', 'Unnamed')
		value = elem.get('value', 'No Value')

		#allow overriding this package with a custom attribute
		pkg = find_attribute(elem, 'FOOTPRINT', variant, pkg)
		desc = find_attribute(elem, 'DESCRIPTION', variant, None)
		pop_attr = find_attribute(elem, 'POPULATE', variant, "yes")

		if pop_attr.lower() == "no":
			return None, True
		elif pop_attr != "yes":
			raise ValueError("Unknown value in POPULATE attribute in element %s: %s" % (name, pop_attr))

		#Only keep the value for part types where that is meaningful
		if not Part.ref.has_value(name):
			value = None

		(manu, mpn) = find_mpn(elem, variant)
		digipn = find_digipn(elem, variant)

		if (mpn is not None and manu is not None) or (digipn is not None and digipn != ""):
			return Part(name, pkg, digipn=digipn, mpn=mpn, manu=manu, value=value, desc=desc, pkg_info=pkg_info), True
		elif mpn == "" or digipn == "":
			return None, True

		return None, False

	def __init__(self, name, package, mpn=None, manu=None, digipn=None, value=None, desc=None, pkg_info=None):
		"""
		Create a part object from the data passed
		"""

		self.name = name
		self.package = package
		self.mpn = mpn
		self.manu = manu
		self.value = value
		self.digipn = digipn
		self.desc = desc
		self.pkg_info = pkg_info

		#If no description is given try creating a generic one based on the type of the part (resistor, etc)
		if self.desc is None:
			self.desc = Part.ref.find_description(self.name, self.value)

	def _parse_type(self, type):
		refs = type.split(',')
		alpha_pat = re.compile('[a-zA-Z]+')

		prefix = re.match(alpha_pat, refs[0])

		if prefix is None:
			return "unknown"

	def unique_id(self):
		"""
		Return a unique key that can be used to group multiple parts that are identical.
		"""

		if self.manu and self.mpn:
			return "%s_%s" % (self.manu, self.mpn)

		return "Digikey_%s" % self.digipn

	def package_info(self):
		if self.pkg_info is None:
			return {'pins': 0, 'pads': 0}

		return {'pins': self.pkg_info.num_pins, 'pads': self.pkg_info.num_pads}

def attrib_name(name, variant):
	if variant == "MAIN" or variant == "":
		return name

	return name + '-' + variant

def find_attribute(part, name, variant, default=None):
	"""
	Find the attribute corresponding the given variant, or if that doesn't exist, 
	the value corresponding to MAIN, otherwise return the given default
	"""

	attrib_var = part.find("./attribute[@name='%s']" % attrib_name(name, variant))
	attrib_main = part.find("./attribute[@name='%s']" % attrib_name(name, 'MAIN'))

	if attrib_var is not None:
		return attrib_var.get('value')

	if attrib_main is not None:
		return attrib_main.get('value')

	return default

def find_mpn(part, variant):
	"""
	See if this part has the variables MPN[-variant] and MANU[-variant] defined
	return None if either is undefined.  Otherwise return the tuple (manu, mpn)
	"""

	mpn = part.find("./attribute[@name='%s']" % attrib_name('MPN', variant))
	manu = part.find("./attribute[@name='%s']" % attrib_name('MANU', variant))

	if mpn is None or manu is None:
		return (None, None)

	return (manu.get('value', 'Unknown'), mpn.get('value', 'Unknown'))

def find_digipn(part, variant):
	pn_elem = part.find("./attribute[@name='%s']" % attrib_name('DIGIKEY-PN', variant))

	if pn_elem is None:
		return None

	return pn_elem.get('value', "")


def convert(arg):
	if arg is None:
		return None

	return str(arg)

def validate_list(arg, choices):
	"""
	Make sure the argument is in the list of choices passed to the function
	"""
	
	choice_set = set(choices)

	if arg not in choices:
		raise ValueError('Value not in list: %s' % str(choices))


#board.py

import csv
import itertools
from part import Part
from xml.etree import ElementTree
from datetime import date
from octopart.identifier import PartIdentifier
from octopart.octopart import Octopart
from reference import PCBReferenceLibrary
from package import Package

class Board:
	def __init__(self, name, file, variants, partname, width, height, revision, unknown, nopop):
		"""
		Create a Board Object with 1 or more assembly variants from the variant dictionary passed in.
		"""

		self.name = name
		self.file = file
		self.variants = {varname: self._process_variant(var) for (varname,var) in variants.iteritems()}
		self.partname = partname
		self.width = width
		self.height = height
		self.revision = revision
		self.unknown = unknown
		self.no_pop = nopop

	def list_variants(self):
		print "Assembly Variants"

		for k in self.variants.iterkeys():
			print k

	def get_ids(self, variant=None):
		"""
		Return a set containing all of the reference identifiers (C1, R7, etc.) in
		the variant passed, or in all variants if no variant is passed
		"""

		if variant is None:
			vars = self.variants.keys()
		else:
			vars = [variant]

		ids = set()

		for var in vars:
			for line in self.variants[var]:
				refs = map(lambda x: x.name, line)
				ids.update(refs)

		#If we're asked for all of the ids on this board, also include those that
		#may never be populated
		if variant is None:
			ids.update(self.no_pop)

		return ids

	def print_variant(self, key):
		var = self.variants[key]

		print "Variant", key
		for v in var:
			print map(lambda x: x.name, v)

	def get_unique_parts(self):
		parts = {}

		for v in self.variants.iterkeys():
			for line in self.variants[v]:
				part = PartIdentifier(line[0])
				parts[part.build_reference()] = part

		return parts

	def lookup_parts(self):
		parts = self.get_unique_parts()

		op = Octopart()

		self.oracle = op.match_identifiers(parts.values())

	def price_variant(self, variant, multiplier, requirements=None):
		if variant not in self.variants:
			raise ValueError("Invalid variant passed to price_variant")

		self.lookup_parts()
		unmatched = []

		prices = []

		for line in self.variants[variant]:
			id = PartIdentifier(line[0]).build_reference()

			if id not in self.oracle:
				unmatched.append(line)
				continue

			quant = int(len(line)*multiplier + 0.5) #round up
			unit_price = self.oracle[id].best_price(quant, requirements)

			if unit_price is None:
				unmatched.append(line)
				continue

			prices.append((line, unit_price))

		return (prices, unmatched)

	def export_list(self, variant):
		"""
		Return an array with all of the digikey part numbers in this variant so that
		it can be concatenated and used to order parts
		"""

		export = []
		ignored = []

		for line in self.variants[variant]:
			pn = line[0].digipn
			if pn is None or pn == "":
				ignored.append(line[0].name)
				continue

			export += [pn]*len(line)

		print "Ignoring Parts"
		print ignored

		return export

	def variant_data(self, variant, include_costs=False, cost_quantity=1, requirements=None):
		"""
		Return a dictionary describing this variant, suitable for using to fill a template.
		Create an array of BOM line items corresponding to this assembly variant
		"""

		res = {}

		if include_costs:
			self.lookup_parts()
			res['include_costs'] = True
			res['cost_quantity'] = cost_quantity
		else:
			res['include_costs'] = False

		var = self.variants[variant]
		lib = PCBReferenceLibrary()
		today = date.today()

		res['company'] = "WellDone International"
		res['part'] = self.partname
		res['variant'] = variant
		res['revision'] = self.revision
		res['date'] = today.strftime('%x')
		res['lines'] = []
		for line in var:
			num = len(line)
			refs = map(lambda x: x.name, line)
			foot = lib.find_package(line[0].package)[0]
			value = line[0].value
			manu = line[0].manu
			mpn = line[0].mpn
			distpn = line[0].digipn
			dist = ""
			descr = line[0].desc
			if distpn:
				dist = "Digikey"

			unitprice = ""
			lineprice = ""

			if include_costs:
				quantity = num*cost_quantity
				id = PartIdentifier(line[0])
				if id.build_reference() in self.oracle:
					price = self.oracle[id.build_reference()].best_price(quantity, requirements)
					if price is not None:
						unitprice = price[0]
						lineprice = num*unitprice

			ldict = {}
			ldict['qty'] = num
			ldict['refs'] = refs
			ldict['footprint'] = foot
			ldict['value'] = value
			ldict['manu'] = manu
			ldict['mpn'] = mpn
			ldict['dist'] = dist
			ldict['description'] = descr
			ldict['distpn'] = distpn
			ldict['unitprice'] = unitprice
			ldict['lineprice'] = lineprice 
			ldict['pkg_info'] = line[0].pkg_info
			
			res['lines'].append(ldict)

		return res

	def export_bom(self, variant, stream, include_costs=False, cost_quantity=1, requirements=None):
		var = self.variants[variant]

		lib = PCBReferenceLibrary()

		lineno = 1

		today = date.today()

		if include_costs:
			self.lookup_parts()

		if isinstance(stream, str):
			bom = open(stream, "wb")
			close = True
		else:
			bom = stream
			close = False
		
		writ = csv.writer(bom)

		writ.writerow(["WellDone"])
		writ.writerow(["BOM: %s (%s)" % (self.partname, variant)])
		writ.writerow(["Revision: %s" % self.revision])
		writ.writerow(['Date Created: %s' % (today.strftime('%x'))])

		if include_costs:
			writ.writerow(['Costs calculated for %d units' % cost_quantity])

		writ.writerow([])
		writ.writerow([])

		headers = ["Item", "Qty", "Reference Design", "Value", "Footprint", "Description", "Manufacturer", "Manu. Part", "Distributor", "Dist. Part"]
		if include_costs:
			headers += ['Unit Price', 'Line Price']

		writ.writerow(headers)

		for line in var:
			num = len(line)
			refs = ", ".join(map(lambda x: x.name, line))
			foot = lib.find_package(line[0].package)[0]
			value = line[0].value
			manu = line[0].manu
			mpn = line[0].mpn
			distpn = line[0].digipn
			dist = ""
			descr = line[0].desc
			if distpn:
				dist = "Digikey"

			row = [lineno, num, refs, value, foot, descr, manu, mpn, dist, distpn]

			if include_costs:
				quantity = num*cost_quantity
				id = PartIdentifier(line[0])
				if id.build_reference() in self.oracle:
					price = self.oracle[id.build_reference()].best_price(quantity, requirements)
					if price is not None:
						unitprice = price[0]
						lineprice = num*unitprice

						row += [str(unitprice), str(lineprice)]

			writ.writerow(row)

			lineno += 1

		#If there are unpopulated parts, list them here
		unpopulated = self.get_ids() - self.get_ids(variant)

		if len(unpopulated) > 0:
			writ.writerow([])

			s = ", ".join(unpopulated)
			writ.writerow(["Unpopulated Parts", s])

		if close:
			bom.close()

	@staticmethod
	def FromEagle(brd_file):
		"""
		Create a Board from EAGLE Board file by parsing attribute tags embedded with the
		components.  Create different assembly variants for each different configuration
		specified in the file
		"""
		
		tree = ElementTree.parse(brd_file)
		if tree is None:
			raise ValueError("File %s is not a valid XML file" % brd_file)

		root = tree.getroot()

		#Find all of the defined packages in this file
		packages = {x.name: x for x in map(lambda x: Package(x), root.findall('.//package'))}

		elems = root.find("./drawing/board/elements")
		if elems is None:
			raise ValueError("File %s does not contain a list of elements.  This is an invalid board file" % brd_file)

		parts = list(elems)

		unknown = []
		nopop = []
		constant = []
		variable = []
		variants = find_variants(root)

		all_vars = set(variants)

		#Create a part object for each part that has a valid digikey-pn
		#If there are multiple digikey part numbers, create a part for each one
		for part in parts:			
			valid_part = False
			found_vars = set()
			nopop_vars = set()

			for v in variants:
				p,found = Part.FromBoardElement(part, v, packages)
				if p and found:
					variable.append( (v, p) )
					found_vars.add(v)
					valid_part = True
				elif found:
					#If this part has digipn == "" or Populate=no, then it should not be populated in this variant
					valid_part = True
					nopop_vars.add(v)
					nopop.append(part.get('name',"Unknown Name"))

			#See if this is a constant part (with no changes in different variants)
			p,found = Part.FromBoardElement(part, "", packages)
			if p and len(found_vars) == 0 and len(nopop_vars) == 0:
				constant.append(p)
				valid_part = True
			elif p and len(found_vars) != len(all_vars):
				#there are some variants that want the default part and some that want a change
				for var in (all_vars - found_vars) - nopop_vars:
					variable.append( (var, p) )
					valid_part = True

			#Make sure this part has information for at least one assembly variant
			if not valid_part:
				unknown.append(part.get('name',"Unknown Name"))

		vars = {}
		#Create multiple variants
		for var in variants:
			vars[var] = constant + filter_sublists(variable, var)

		return Board("test", 'test', vars,
						partname=find_attribute(root, 'PARTNAME'),
						width=find_attribute(root, 'WIDTH'),
						height=find_attribute(root, 'HEIGHT'),
						revision=find_attribute(root, 'REVISION'),
						unknown=unknown,
						nopop=nopop)

	def _process_variant(self, parts):
		"""
		Group items by unique key and return tuples of identical parts
		"""

		bomitems = []

		sparts = sorted(parts, key=lambda x:x.unique_id())
		for k,g in itertools.groupby(sparts, lambda x:x.unique_id()):
			bomitems.append(list(g))

		return bomitems


def remove_prefix(s, prefix):
	if not s.startswith(prefix):
		return s

	return s[len(prefix):]

def filter_sublists(master, key):
	"""
	Given a list of lists where each of the sublist elements are tuples (key, value),
	return a concatenated list of all values in tuples that match key
	"""

	concat = itertools.chain(master)

	filtered = filter(lambda x: x[0] == key, concat)

	return map(lambda x:x[1], filtered)

def get_variant_id(pn_name):
	known_prefixes = ['DIGIKEY-PN']

	for p in known_prefixes:
		pn_name = remove_prefix(pn_name, p)

	if pn_name.startswith('-'):
		pn_name = pn_name[1:]

	return pn_name

def find_attribute(root, name):
	attr = root.find(".//attribute[@name='%s']" % name)
	if attr is None:
		raise ValueError("Required global board attribute not found: %s" % name)

	return attr.get('value')

def find_variants(root):
	vars = root.findall(".//attribute[@value='%s']" % 'ASSY-VARIANT')

	if len(vars) == 0:
		return ["MAIN"]
	else:
		return map(lambda x: x.get('name'), vars)


from setuptools import setup, find_packages
setup(
    name = "pymomo",
    packages = find_packages(),
    version = "1.0.0",
    install_requires=[
        "beautifulsoup4==4.3.2",
        "BTrees==4.1.1",
        "Cheetah==2.4.4",
        "cmdln==1.1.2",
        "colorama==0.3.3",
        "decorator==3.4.0",
        "dirspec==13.10",
        "intelhex==1.5",
        "Markdown==2.5.2",
        "nose==1.3.4",
        "persistent==4.0.8",
        "py==1.4.26",
        "pycparser==2.10",
        "pyparsing==2.0.3",
        "pyserial==2.7",
        "pytest==2.6.4",
        "six==1.9.0",
        "transaction==1.4.3",
        "zc.lockfile==1.1.0",
        "ZConfig==3.0.4",
        "zdaemon==4.0.1",
        "ZEO==4.1.0",
        "ZODB==4.1.0",
        "ZODB3==3.11.0",
        "zope.interface==4.1.2"
    ],
    package_data={
        'pymomo': ['config/*']
    },
    description = "WellDone Mobile Monitor (MoMo) Interaction Library",
    author = "WellDone International",
    author_email = "info@welldone.org",
    url = "http://github.com/welldone/pymomo",
    keywords = ["momo", "remote", "monitoring", "embedded", "hardware"],
    classifiers = [
        "Programming Language :: Python",
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description = """\
PyMoMo
------

A python library for interacting with MoMo devices.  See https://momo.welldone.org.

"""
)


#prices.py
#An object encapsulating a list of pricebreaks

from decimal import Decimal

class PriceList:
	def __init__(self, octo_pricelist):
		"""
		Create from an octopart PriceList json object
		Assume that we want to work in USD
		"""

		self.breaks = []

		if 'USD' not in octo_pricelist:
			raise ValueError('USD Not in Price Object')

		plist = octo_pricelist['USD']

		for i in xrange(0, len(plist)):
			q,val = plist[i]
			price = Decimal(val)
			self.breaks.append((q, price))

	def unit_price(self, quantity):
		if len(self.breaks) == 0:
			return None

		for i in xrange(0, len(self.breaks)):
			if self.breaks[i][0] > quantity:
				if i == 0:
					return None

				return self.breaks[i-1][1]

		return self.breaks[-1][1]

	def total_price(self, quantity):
		up = self.unit_price(quantity)

		if up is None:
			return None

		return quantity*up


#physicalpart.py
#A class representing a Part in the Octopart API

from partoffer import PartOffer
import utilities

class PhysicalPart:
	def __init__(self, response):
		utilities.assert_class(response, 'Part')

		self.offers = map(lambda x: PartOffer(x), response['offers'])

	def best_price(self, quantity, requirements):
		valid_offers = filter(lambda x: requirements.validate(x, quantity), self.offers)
		prices = map(lambda x: (x.best_price(quantity), x.seller_name, x.packaging), valid_offers)		
		valid_prices = filter(lambda x: x[0] is not None, prices)

		sorted_prices = sorted(valid_prices, key=lambda x:x[0])

		if len(sorted_prices) > 0:
			return sorted_prices[0]

		return None